@@ -2,7 +2,7 @@ package CPAN::HandleConfig;
 use strict;
 use vars qw(%can %keys $dot_cpan $VERSION);
 
-$VERSION = sprintf "%.2f", substr(q$Rev: 423 $,4)/100;
+$VERSION = sprintf "%.2f", substr(q$Rev: 485 $,4)/100;
 
 %can = (
   'commit' => "Commit changes to disk",
@@ -14,7 +14,7 @@ $VERSION = sprintf "%.2f", substr(q$Rev: 423 $,4)/100;
     build_cache build_dir bzip2
     cache_metadata commandnumber_in_prompt cpan_home curl
     dontload_hash
-    ftp ftp_proxy
+    ftp ftp_passive ftp_proxy
     getcwd gpg gzip
     histfile histsize http_proxy
     inactivity_timeout index_expire inhibit_startup_message
@@ -29,6 +29,10 @@ $VERSION = sprintf "%.2f", substr(q$Rev: 423 $,4)/100;
     unzip urllist
     wait_list wget
 );
+if ($^O eq "MSWin32") {
+    delete $keys{mbuild_install_build_command};
+    delete $keys{make_install_make_command};
+}
 
 # returns true on successful action
 sub edit {
@@ -79,7 +83,11 @@ sub edit {
                 undef $CPAN::FTP::Themethod;
             }
             return $changed;
-	} else {
+        } elsif ($o =~ /_hash$/) {
+            @args = () if @args==1 && $args[0] eq "";
+            push @args, "" if @args % 2;
+            $CPAN::Config->{$o} = { @args };
+        } else {
 	    $CPAN::Config->{$o} = $args[0] if defined $args[0];
 	    $self->prettyprint($o);
 	}
@@ -159,9 +167,14 @@ EOF
         $CPAN::Frontend->mydie("Couldn't open >$configpm: $!");
     $fh->print(qq[$msg\$CPAN::Config = \{\n]);
     foreach (sort keys %$CPAN::Config) {
+        unless (exists $keys{$_}) {
+            $CPAN::Frontend->mywarn("Dropping unknown config variable '$_'\n");
+            delete $CPAN::Config->{$_};
+            next;
+        }
 	$fh->print(
 		   "  '$_' => ",
-		   ExtUtils::MakeMaker::neatvalue($CPAN::Config->{$_}),
+		   $self->neatvalue($CPAN::Config->{$_}),
 		   ",\n"
 		  );
     }
@@ -176,11 +189,40 @@ EOF
     1;
 }
 
-*default = \&defaults;
+# stolen from MakeMaker; not taking the original because it is buggy;
+# bugreport will have to say: keys of hashes remain unquoted and can
+# produce syntax errors
+sub neatvalue {
+    my($self, $v) = @_;
+    return "undef" unless defined $v;
+    my($t) = ref $v;
+    return "q[$v]" unless $t;
+    if ($t eq 'ARRAY') {
+        my(@m, @neat);
+        push @m, "[";
+        foreach my $elem (@$v) {
+            push @neat, "q[$elem]";
+        }
+        push @m, join ", ", @neat;
+        push @m, "]";
+        return join "", @m;
+    }
+    return "$v" unless $t eq 'HASH';
+    my(@m, $key, $val);
+    while (($key,$val) = each %$v){
+        last unless defined $key; # cautious programming in case (undef,undef) is true
+        push(@m,"q[$key]=>".$self->neatvalue($val)) ;
+    }
+    return "{ ".join(', ',@m)." }";
+}
+
 sub defaults {
     my($self) = @_;
-    $self->unload;
-    $self->load;
+    my $done;
+    for my $config (qw(CPAN/MyConfig.pm CPAN/Config.pm)) {
+      CPAN::Shell->reload_this($config) and $done++;
+      last if $done;
+    }
     1;
 }
 
@@ -255,20 +297,26 @@ sub load {
 	my($path_to_cpan) = File::Basename::dirname($INC{"CPAN.pm"});
 	my($configpmdir) = File::Spec->catdir($path_to_cpan,"CPAN");
 	my($configpmtest) = File::Spec->catfile($configpmdir,"Config.pm");
+        my $inc_key;
 	if (-d $configpmdir or File::Path::mkpath($configpmdir)) {
-	    $configpm = _configpmtest($configpmdir,$configpmtest); 
+	    $configpm = _configpmtest($configpmdir,$configpmtest);
+            $inc_key = "CPAN/Config.pm";
 	}
 	unless ($configpm) {
 	    $configpmdir = File::Spec->catdir($ENV{HOME},".cpan","CPAN");
 	    File::Path::mkpath($configpmdir);
 	    $configpmtest = File::Spec->catfile($configpmdir,"MyConfig.pm");
-	    $configpm = _configpmtest($configpmdir,$configpmtest); 
-	    unless ($configpm) {
-			my $text = qq{WARNING: CPAN.pm is unable to } .
-			  qq{create a configuration file.}; 
-			output($text, 'confess');
-	    }
+	    $configpm = _configpmtest($configpmdir,$configpmtest);
+            $inc_key = "CPAN/MyConfig.pm";
 	}
+        if ($configpm) {
+          $INC{$inc_key} = $configpm;
+        } else {
+          my $text = qq{WARNING: CPAN.pm is unable to } .
+              qq{create a configuration file.};
+          output($text, 'confess');
+        }
+
     }
     local($") = ", ";
     $CPAN::Frontend->myprint(<<END) if $redo && ! $theycalled;
@@ -314,16 +362,12 @@ sub missing_config_data {
          "unzip",
          "urllist",
         ) {
+        next unless exists $keys{$_};
 	push @miss, $_ unless defined $CPAN::Config->{$_};
     }
     return @miss;
 }
 
-sub unload {
-    delete $INC{'CPAN/MyConfig.pm'};
-    delete $INC{'CPAN/Config.pm'};
-}
-
 sub help {
     $CPAN::Frontend->myprint(q[
 Known options:
@@ -385,7 +429,7 @@ package ####::###### #hide from indexer
 
 use strict;
 use vars qw($AUTOLOAD $VERSION);
-$VERSION = sprintf "%.2f", substr(q$Rev: 423 $,4)/100;
+$VERSION = sprintf "%.2f", substr(q$Rev: 485 $,4)/100;
 
 # formerly CPAN::HandleConfig was known as CPAN::Config
 sub AUTOLOAD {
@@ -400,5 +444,5 @@ sub AUTOLOAD {
 __END__
 # Local Variables:
 # mode: cperl
-# cperl-indent-level: 2
+# cperl-indent-level: 4
 # End:
